unit GameUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, MMSystem, MainUnit, NameUnit;

{ Konstanty }

{ Konstanty zavisejici nejakym zpusobem na absolutnim smeru pohybu hada - vzdy
  velikost 0..3, protoze existuji prave 4 ruzne smery pohybu 0...Nahoru,
  1...Dolu, 2...Doleva, 3...Doprava }
{ Tabulky, ktere rikaji, jaka cast tela hada se nove prida na misto krku,
  prvni index udava novy smer, druhy index minuly smer pohybu . Prvni
  (resp. druha) tabulka zjistuje levou (resp. pravou) hranici obdelniku
  pro vyriznuti obrazku teto cati z HiddenImage. }
const NibbleNecksLeft: array[0..3, 0..3] of Byte =
       ( ( 4*15,    0, 2*15, 3*15 ),
         (    0, 4*15, 0*15, 1*15 ),
         ( 1*15, 3*15, 5*15,    0 ),
         ( 0*15, 2*15,    0, 5*15 ) );
const NibbleNecksRight: array[0..3, 0..3] of Byte =
       ( ( 5*15,    0, 3*15, 4*15 ),
         (    0, 5*15, 1*15, 2*15 ),
         ( 2*15, 4*15, 6*15,    0 ),
         ( 1*15, 3*15,    0, 6*15 ) );
{ Pole obsahujici prirustky X-ove a Y-ove souradnice v zavislosti na
  smeru pohybu. }
const DirectionX: array[0..3] of Shortint = ( 0, 0, -1, 1 );
const DirectionY: array[0..3] of Shortint = ( -1, 1, 0, 0 );
{ Pole obsahujici novy smer pohybu pri 2-klavesovem ovladani,
  index znaci predchozi smer pohybu. }
const DirectionLeft: array[0..3] of Byte = ( 2, 3, 1, 0);
const DirectionRight: array[0..3] of Byte = ( 3, 2, 0, 1);

{ Maximalni indexy v hraci plose ve ctvercich 15x15 pixelu (v podstate
  velikost hraci plochy - 1), mely by odpovidat "fyzicke" velikosti GameImage }
const MaxXIndex = 31;
const MaxYIndex = 23;

{ Maximalni index smeru hada a jejich celkovy pocet }
const MaxDirection = 3;
const TotDirections = MaxDirection + 1;

{ Stavy hada }
const SDeath = 0; { Had zemre }
const SBonus = 1; { Sezere bonus }
const SFree  = 2; { Jede na volne policko }
const STail  = 3; { Jede na policko s ocasem nektereho hada }

{ Jmena zvukovych souboru }
const BonusWave: array[0..4] of PChar = ('Bonus1.wav', 'Bonus2.wav', 'Bonus3.wav', 'Bonus4.wav', 'Bonus5.wav');
const HighScoreWave: array[0..2] of PChar = ('Score1.wav', 'Score2.wav', 'Score3.wav' );
const GameOverWave: array[0..1] of PChar = ('GamOver1.wav', 'GamOver2.wav' );
const DeathWave: array[0..2] of PChar = ('Death1.wav', 'Death2.wav', 'Death3.wav' );

type
  TGameForm = class(TForm)
    GameImage: TImage;
    Timer1: TTimer;
    RedScoreLabel: TLabel;
    GreenScoreLabel: TLabel;
    YellowScoreLabel: TLabel;
    BlueScoreLabel: TLabel;
    HiddenImage: TImage;
    GameOverLabel: TLabel;
    PauseButton: TButton;
    TitleLabel: TLabel;
    Bevel2: TBevel;
    procedure Timer1Timer(Sender: TObject);
    procedure FormActivate(Sender: TObject);
    procedure PauseButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    { Matice s hraci plochou, delka prvku je 32 bitu, kazdy ma nasledujici vyznam:
        1.-2. nejvyznamnejsi bit: Barva hada (0=Blue, 1=Yellow, 2=Green, 3=Red)
        3.-4. nejvyznamnejsi bit: Smer, kterym had z tohoto policka pokracoval
          na sousedni policko (0...Nahoru, 1...Dolu, 2...Doleva, 3...Doprava).
          Pro hlavu hada neni samozrejme definovan.
        5. nejvyznamnejsi bit: Vyuzit pri generovani tahu pocitace, je-li 1,
          pozice uz byla prohledavana - viz TGameForm.FindDirectionToSurvive.
        24 nejmene vyznamnych bitu: Cas, kdy dane policko bylo zabrano telem
          hada, mereno v poctech tahu od zacatku hry. Pokud je tento cas
          roven 0, policko neni obsazeno zadnym hadem (barva a smer musi byt take 0). }
    { Matice ma o jednicku na kazde strane vetsi rozmery, aby nedoslo k prekroceni
      mezi pri nekterych operacich. }
    Desk: array[-1..MaxXIndex+1, -1..MaxYIndex+1] of Longword;
    { Pole s informacemi o hadech - pro kazdeho jeden record, opet neni kvuli
      prekladaci zapsana konstanta primo }
    Nibble: array[0..MaxColor] of record
      { Souradnice hlav, ocasu a casti tela pred ocasem hadu mereno
        ve hracich polich }
      HeadX, HeadY,
      TailX, TailY,
      BeforeTailX, BeforeTailY: Shortint;
      { Aktualni skore hadu }
      Score: Longint;
      { Delka hada v hracich polich, je-li 0, had neni pritomen }
      Length: Word;
      { Aktualni a novy smer hada }
      Direction, NewDirection: Byte;
      { Jestli se jedna o lidskeho nebo pocitacoveho hrace }
      IsHuman: Boolean;
      { Jestli hraje hrac dvema nebo ctyrma klavesama }
      UsesTwoKeys: Boolean;
      { Pocet tahu, ktere had promysli dopredu = poc. inteligence }
      MaxDepth: Byte;
      { Stav hada po volani funkci PrepareNibbleTurn nebo DoNibbleTurn,
        0=Smrt, 1=Sezral bonus, 2=Jede na neobsazene pole, 3=Jede na ocas }
      State: Byte;
      { Nove souradnice hlavy po volani fce PrepareNibbleTurn }
      NewHeadX, NewHeadY: Shortint;
      { Barva hada, na jehoz ocas had najizdi. Je to jen k urychleni vypoctu. }
      TailColor: Byte;
      { Na kterych pocatecnich pozicich uz byli vytvoreni novy hadi,
        nejedna se o vlastnost jednotliveho hada, ale pocatecnich pozic je
        stejne jako hadu, tedy muzeme to umistit sem. }
      PositionOccupied: Boolean;
      { Virtual key codes klaves ovladani, poradi clenu: 0...Nahoru/otoceni doleva,
        1...Dolu/otoceni doprava, 2...Doleva, 3...Doprava }
      KeySettings: array[0..MaxDirection] of Byte;
    end;
    { Souradnice bonusu na hraci plose - vzdy existuje prave jeden }
    BonusX, BonusY: Shortint;
    { O kolik se zvysi score hada kdyz sezere bonus - zavisi na inteligenci
      a poctu vsech hadu a rychlosti hry }
    ScoreAdd: Word;
    { Cislo aktualniho tahu }
    dwGameTurn: Longword;
    { Dalsi pomocne promenne }
    bBonusCount: Byte;          { Pocet hadu, kteri v danem tahu sezerou bonus }
    fGenerateNewBonus: Boolean; { Jestli se bude generovat novy bonus }
    fGameOver: Boolean;         { Jestli je konec hry }
    fPaused: Boolean;           { Jestli je hra zastavena }
    fRefreshScore: Boolean;     { Jestli se ma prepsat skore }
    fPlaySound: Boolean;        { Jestli se bude prehravat zvuk }
    InitKeyboardState: TKeyboardState; { Uvodni stav klavesnice }

    { Procedury a funkce }
    procedure CreateNibble(bColor: Byte);
    procedure PrepareNibbleTurn(bColor: Byte);
    procedure DoNibbleTurn(bColor: Byte);
    procedure GenerateNewBonus;
    procedure RefreshScore;
    procedure DoNibblesDeath;
    procedure DoGameOver;
    function GenerateNibbleTurn(bColor: Byte): Byte;
    function FindDirectionToSurvive(bDepth, bColor, bDirection: Byte; sX, sY: Shortint): Byte;

  public
    { Public declarations }
  end;

var
  GameForm: TGameForm;

implementation

{$R *.DFM}


{ Procedura nastavajici pri aktivaci formy }
procedure TGameForm.FormActivate(Sender: TObject);
var
  bCounter1, bCounter2, bColor: Byte;
begin
  { Takoveto "pozadi" bude mit hra }
  GameImage.Canvas.Brush.Bitmap := nil;
  GameImage.Canvas.Brush.Style := bsSolid;
  GameImage.Canvas.Brush.Color := clBlack;
  GameImage.Canvas.FillRect( Rect(0, 0, GameImage.Width, GameImage.Height) );
  GameOverLabel.Caption := '';
  TitleLabel.Caption := '';

  { Inicializace nahodneho generatoru }
  Randomize;

  { Nastavime rychlost timeru v GameForm podle trackbaru }
  Timer1.Interval := 555 - (5 * MainForm.GameSpeedTrackBar.Position);

  { Nastavime uvodni stav klavesnice - vsechny klavesy vypnuty, pouze
    NumLock zustava neustale ve stavu, v jakem byl pri aktivaci teto formy,
    protoze jinak by ho na nekterych pocitacich funkce SetKeyboardState
    vypinala. }
  ZeroMemory(@InitKeyboardState, SizeOf(InitKeyboardState));
  InitKeyboardState[VK_NUMLOCK] := GetKeyState(VK_NUMLOCK);
  SetKeyboardState(InitKeyboardState);

  { "Stahneme" si uzivatelske nastaveni klaves }
  for bColor := 0 to MaxColor do
    for bCounter2 := 0 to MaxPanel do
      Nibble[bColor].KeySettings[bCounter2] := MainForm.KeySettings[bColor, bCounter2];

  { Inicializace promennych }
  GameImage.Canvas.CopyMode := cmSrcCopy;
  ZeroMemory(@Desk, SizeOf(Desk));
  for bColor := 0 to MaxColor do
  begin
    Nibble[bColor].Score := 0;
    Nibble[bColor].Length := 0;
    Nibble[bColor].IsHuman := false;
    Nibble[bColor].UsesTwoKeys := false;
    Nibble[bColor].HeadX := -1;
    Nibble[bColor].HeadY := -1;
    Nibble[bColor].State := SDeath;
    Nibble[bColor].NewHeadX := -1;
    Nibble[bColor].NewHeadY := -1;
    Nibble[bColor].PositionOccupied := false;
  end;
  fPaused := false;
  fGameOver := false;
  fGenerateNewBonus := true;
  fRefreshScore := false;
  fPlaySound := MainForm.SoundCheckBox.Checked;

  { Nastavime ovladani - 2/4 klavesy }
  for bColor := 0 to MaxColor do
    if MainForm.NibbleInterface[bColor].KeySpeedButton.Down then
      Nibble[bColor].UsesTwoKeys := true;

  { Vytvorime hady a spocitame pocet "lidskych" a "pocitacovych" hadu }
  bCounter1 := 0; { Pocet "lidskych" }
  bCounter2 := 0; { Soucet inteligence "pocitacovych" hadu }
  for bColor := 0 to MaxColor do
  if MainForm.NibbleInterface[bColor].CheckBox.Checked then
  begin
    CreateNibble(bColor);
    if MainForm.NibbleInterface[bColor].HumSpeedButton.Down then
    begin
      Inc(bCounter1);
      Nibble[bColor].IsHuman := true;
    end
    else
      Inc(bCounter2, MainForm.NibbleInterface[bColor].IQTrackBar.Position);
  end;

  { Vypocitame prirustek skore podle jisteho vzorce }
  ScoreAdd := 101 + (bCounter1 * 11) + (bCounter2 * 2);
  ScoreAdd := ScoreAdd + (ScoreAdd * MainForm.GameSpeedTrackBar.Position div 50);

  { Vygenerujeme novy bonus }
  GenerateNewBonus;

  { Nastavime pocitacovou inteligenci, MaxDepth by mela byt vzdy mensi
    nez 34 - viz GameDesk.FindDirectionAndSurvive }
  for bColor := 0 to MaxColor do
    Nibble[bColor].MaxDepth := 3 + MainForm.NibbleInterface[bColor].IQTrackBar.Position * 3;

  { A spustime timer }
  Timer1.Enabled := true;
end;

{ Takt timeru }
procedure TGameForm.Timer1Timer(Sender: TObject);
var
  KeyboardState: TKeyboardState;
  bHelpValue: Byte;
  bColor: Byte;
begin
  { Inicializace }
  dwGameTurn := dwGameTurn + 1;
  bBonusCount := 0;

  { Zjistime novy stav klavesnice }
  GetKeyboardState(KeyboardState);

  { Nyni ze stavu klavesnice odvodime smery pohybu a pripravime tahy }
  for bColor := 0 to MaxColor do
  begin
    { Hraje had vubec ? }
    if Nibble[bColor].Length = 0 then
      Continue;

    { Hraje-li za hada pocitac, vygenerujeme tah }
    if not Nibble[bColor].IsHuman then
      Nibble[bColor].NewDirection := GenerateNibbleTurn(bColor)

    else { jinak tah zjistime z klavesnice }

      { 2-klavesove ovladani }
      if Nibble[bColor].UsesTwoKeys then

        if KeyboardState[ Nibble[bColor].KeySettings[TurnLeft] ] > 0 then
          Nibble[bColor].NewDirection := DirectionLeft[ Nibble[bColor].Direction ]
        else
          if KeyboardState[ Nibble[bColor].KeySettings[TurnRight] ] > 0 then
            Nibble[bColor].NewDirection := DirectionRight[ Nibble[bColor].Direction ]
          else
            Nibble[bColor].NewDirection := Nibble[bColor].Direction

      else

        { 4-klavesove ovladani }
        if KeyboardState[ Nibble[bColor].KeySettings[GoUp] ] > 0 then
          Nibble[bColor].NewDirection := GoUp { Nahoru }
        else
          if KeyboardState[ Nibble[bColor].KeySettings[GoDown] ] > 0 then
            Nibble[bColor].NewDirection := GoDown { Dolu }
          else
            if KeyboardState[ Nibble[bColor].KeySettings[GoLeft] ] > 0 then
              Nibble[bColor].NewDirection := GoLeft { Doleva }
            else
              if KeyboardState[ Nibble[bColor].KeySettings[GoRight] ] > 0 then
                Nibble[bColor].NewDirection := GoRight { Doprava }
              else
                Nibble[bColor].NewDirection := Nibble[bColor].Direction;

    { Pokud by jel had sam "proti sve hlave" (cili zemrel by), tak upravime jeho tah tak,
      aby se to nestalo - mala vypomoc lidskym hracum. }
    bHelpValue := Nibble[bColor].NewDirection + Nibble[bColor].Direction;
    if (bHelpValue = 5) or (bHelpValue = 1) then
      Nibble[bColor].NewDirection := Nibble[bColor].Direction;
    { A pripravime tah hada }
    PrepareNibbleTurn(bColor);
  end;

  { Ted tah provedeme - bez for cyklu pro urychleni }
  for bColor := 0 to MaxColor do
    DoNibbleTurn(bColor);

  { A nyni obstareme vsechny pripadne herni udalosti }
  GenerateNewBonus;
  RefreshScore;
  DoNibblesDeath;
  DoGameOver;

  { Nastavime klavesnici do uvodniho stavu }
  SetKeyboardState(InitKeyboardState);

end;

{ Tato procedura vytvori hada bColor (0=Blue, 1=Yellow, 2=Green, 3=Red)
  na nahodne vybrane startovni pozici ze ctyr moznych }
procedure TGameForm.CreateNibble(bColor: Byte);
var
  wPosition, wPosX, wPosY: Smallint;
  bDirection: Byte;
  TileImageRect, DestRect: TRect;
begin
  { Nahodne vybereme volnou pozici pro umisteni hada }
  repeat
    wPosition := Random(TotColors);
  until Nibble[wPosition].PositionOccupied = false;
  Nibble[wPosition].PositionOccupied := true;

  { Mame 4 ruzne moznosti volnych pozic - tedy max. pro ctyri hady }
  case wPosition of
    0: begin wPosX := MaxXIndex div 2;     wPosY := MaxYIndex - 2;       bDirection := 0; end;
    1: begin wPosX := (MaxXIndex div 2)+1; wPosY := 2;                   bDirection := 1; end;
    2: begin wPosX := 2;                   wPosY := MaxYIndex div 2;     bDirection := 3; end;
  else { 3: }wPosX := MaxXIndex - 2;       wPosY := (MaxYIndex div 2)+1; bDirection := 2;
  end;

  { Inicializujeme zakladni promenne }
  dwGameTurn := 3; { = Delka hada }
  Nibble[bColor].Length := 3;
  Nibble[bColor].Direction := bDirection;
  Nibble[bColor].State := SFree;

  { Vytvorime a namalujeme ocas hada }
  Desk[wPosX, wPosY] := (Longword(bColor) shl 30) or (Longword(bDirection) shl 28) or $00000001;
  Nibble[bColor].TailX := wPosX;
  Nibble[bColor].TailY := wPosY;
  TileImageRect := Rect(150 + 15*bDirection, bColor*15, 165 + 15*bDirection, (bColor+1)*15);
  DestRect := Rect(wPosX*15, wPosY*15, (wPosX+1)*15, (wPosY+1)*15);
  GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, TileImageRect);

  { Vytvorime a namalujeme cast tela za ocasem }
  wPosX := wPosX + DirectionX[bDirection];
  wPosY := wPosY + DirectionY[bDirection];
  Desk[wPosX, wPosY] := (Longword(bColor) shl 30) or (Longword(bDirection) shl 28) or $00000002;
  Nibble[bColor].BeforeTailX := wPosX;
  Nibble[bColor].BeforeTailY := wPosY;
  if bDirection <= 1 then
    TileImageRect := Rect(60, bColor*15, 75, (bColor+1)*15)
  else
    TileImageRect := Rect(75, bColor*15, 90, (bColor+1)*15);
  DestRect := Rect(wPosX*15, wPosY*15, (wPosX+1)*15, (wPosY+1)*15);
  GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, TileImageRect);

  { A nakonec vytvorime a namalujeme hlavu }
  wPosX := wPosX + DirectionX[bDirection];
  wPosY := wPosY + DirectionY[bDirection];
  Desk[wPosX, wPosY] := (Longword(bColor) shl 30) or $00000003;
  Nibble[bColor].HeadX := wPosX;
  Nibble[bColor].HeadY := wPosY;
  TileImageRect := Rect(90 + 15*bDirection, bColor*15, 105 + 15*bDirection, (bColor+1)*15);
  DestRect := Rect(wPosX*15, wPosY*15, (wPosX+1)*15, (wPosY+1)*15);
  GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, TileImageRect);

end;


{ Tato procedura pripravi tah hada bColor smerem bDirection (resi hlavne
  ruzne kolize hadu). Tah se muze provest az kdyz jsou takto pripraveny vsechny
  tahy. Parameter bColor urcuje barvu hada: 0=Blue, 1=Yellow, 2=Green, 3=Red. }
procedure TGameForm.PrepareNibbleTurn(bColor: Byte);
var
  bDirection: Byte;
  sNewHeadX, sNewHeadY: Shortint;
  bHelpColor: Byte;
begin
  if Nibble[bColor].Length = 0 then
    Exit; { Tento had vubec nehraje }

  { Stahneme si novy smer pohybu }
  bDirection := Nibble[bColor].NewDirection;

  { Zjistime novou polohu hlavy }
  sNewHeadX := Nibble[bColor].HeadX + DirectionX[bDirection];
  sNewHeadY := Nibble[bColor].HeadY + DirectionY[bDirection];
  Nibble[bColor].NewHeadX := sNewHeadX;
  Nibble[bColor].NewHeadY := sNewHeadY;

  { Nyni priradime hadovi stav }
  { Nenaboural had do zdi ? }
  if (sNewHeadY < 0) or (sNewHeadY > MaxYIndex) or (sNewHeadX < 0) or (sNewHeadX > MaxXIndex) then
    begin Nibble[bColor].State := SDeath; Exit; end;

  { Nesezral had bonus ? }
  if (BonusX = sNewHeadX) and (BonusY = sNewHeadY) then
    begin Nibble[bColor].State := SBonus; Inc(bBonusCount); Exit; end;

  { Nejede had na neobsazene policko ? }
  if Desk[sNewHeadX, sNewHeadY] = 0 then
    begin Nibble[bColor].State := SFree; Exit; end;

  { Ted tedy vime, ze had jede na jiz obsazene policko }
  { Jen tedy zjistime, zda je toto policko ocasem nektereho hada.
    Kdyz bude, pak had nemusi nutne umrit, protoze pokud ten had,
    do jehoz ocasu se "nabouralo" nesezere bonus, posune se v dalsim
    tahu tento jeho ocas. }
  bHelpColor := Desk[sNewHeadX, sNewHeadY] shr 30;
  if (Nibble[bHelpColor].TailX = sNewHeadX) and
     (Nibble[bHelpColor].TailY = sNewHeadY) then
    begin Nibble[bColor].State := STail; Nibble[bColor].TailColor := bHelpColor; Exit; end;

  { Had tedy naboural do tela nejakeho hada (ne ale do ocasu) - smrt }
  Nibble[bColor].State := SDeath;

end;


{ Tato procedura provede tah hada bColor. Tah se muze provest az pote,
  co je pro kazdeho hada zavolana procedura PrepareNibbleTurn.
  Parameter bColor urcuje barvu hada: 0=Blue, 1=Yellow, 2=Green, 3=Red.
  Procedura obstara vsechny udalosti krome smrti a take prekresli hraci plochu. }
procedure TGameForm.DoNibbleTurn(bColor: Byte);
var
  { Pomocne promenne }
  sNewHeadX, sNewHeadY: Shortint;
  sOldHeadX, sOldHeadY: Shortint;
  sOldTailX, sOldTailY: Shortint;
  sOldBeforeTailX, sOldBeforeTailY: Shortint;
  bOldDirection: Byte;
  bDirection: Byte;
  bColorIndex: Byte;
  HeadImageRect, NeckImageRect, TileImageRect, DestRect: TRect;
begin
  { Je vubec had pritomen a ziv ? }
  if (Nibble[bColor].Length = 0) or (Nibble[bColor].State = SDeath) then
    Exit;

  { Naplnime pomocne promenne }
  bDirection := Nibble[bColor].NewDirection;
  sNewHeadX := Nibble[bColor].NewHeadX;
  sNewHeadY := Nibble[bColor].NewHeadY;
  sOldHeadX := Nibble[bColor].HeadX;
  sOldHeadY := Nibble[bColor].HeadY;

  { Nechteji dva hadi vjet na teze policko ? Pokud ano, zemrou oba. }
  for bColorIndex := 0 to MaxColor do
    if (Nibble[bColorIndex].NewHeadX = sNewHeadX) and
       (Nibble[bColorIndex].NewHeadY = sNewHeadY) and
       (bColor <> bColorIndex) then
      begin Nibble[bColorIndex].State := SDeath; Nibble[bColor].State := SDeath; end;

  { Chce tedy had vjet na teze policko jako nejaky jiny ? }
  if Nibble[bColor].State = SDeath then
    Exit;

  { Had bColor chce vjet na policko s ocasem nejakeho hada, tak jestli tento
    sezere bonus a prodluzi se, bColor konci. }
  if Nibble[bColor].State = STail then
    if (Nibble[Nibble[bColor].TailColor].State = SBonus) and (bBonusCount = 1) then
      begin Nibble[bColor].State := SDeath; Exit; end;

  { Ted je jiz vse OK, takze tah hada provedeme }
  Desk[sNewHeadX, sNewHeadY] := (Longword(bColor) shl 30) or dwGameTurn;
  { Pridani 2 smerovych bitu }
  Desk[sOldHeadX, sOldHeadY] := (Longword(bColor) shl 30)
                                or (Longword(bDirection) shl 28)
                                or (dwGameTurn - 1) ;
  Nibble[bColor].HeadX := sNewHeadX;
  Nibble[bColor].HeadY := sNewHeadY;

  if Nibble[bColor].State = SBonus then
  begin { Had sezral bonus - ocas zustava na miste }
    Inc(Nibble[bColor].Length);
    { Po skonceni provadeni tahu se vygeneruje novy bonus }
    fGenerateNewBonus := true;
    { A zvysime trochu skore }
    Inc(Nibble[bColor].Score, ScoreAdd + Random(8));
    fRefreshScore := true;
    { Prehrajeme zvuk }
    if fPlaySound then
      sndPlaySound(BonusWave[Random(5)], SND_ASYNC);
  end
  else
  begin { Had nesezral bonus - ocas se posune }
    sOldTailX := Nibble[bColor].TailX;
    sOldTailY := Nibble[bColor].TailY;
    sOldBeforeTailX := Nibble[bColor].BeforeTailX;
    sOldBeforeTailY := Nibble[bColor].BeforeTailY;
    bOldDirection := (Desk[sOldBeforeTailX, sOldBeforeTailY] shr 28) and $03;

    Nibble[bColor].TailX := sOldBeforeTailX;
    Nibble[bColor].TailY := sOldBeforeTailY;

    Nibble[bColor].BeforeTailX := sOldBeforeTailX + DirectionX[bOldDirection];
    Nibble[bColor].BeforeTailY := sOldBeforeTailY + DirectionY[bOldDirection];

    { Nyni "vymazeme" puvodni policko ocasu z GameImage a Desk, ovsem jen pokud
      si tam jiny had uz nenamaloval hlavu }
    if (Desk[sOldTailX, sOldTailY] and $CFFFFFFF)
       = ((Longword(bColor) shl 30) or (dwGameTurn-Nibble[bColor].Length)) then
    begin
      GameImage.Canvas.FillRect(Rect(sOldTailX*15, sOldTailY*15, (sOldTailX+1)*15, (sOldTailY+1)*15));
      Desk[sOldTailX, sOldTailY] := 0;
    end;

    { A nakreslime ocas }
    TileImageRect := Rect(150 + 15*bOldDirection, bColor*15, 165 + 15*bOldDirection, (bColor+1)*15);
    DestRect := Rect(sOldBeforeTailX*15, sOldBeforeTailY*15, (sOldBeforeTailX+1)*15, (sOldBeforeTailY+1)*15);
    GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, TileImageRect);

  end;

  { Pripravime si hranici obrazku z HiddenImage pro prekresleni krku a hlavy }
  NeckImageRect := Rect(NibbleNecksLeft[bDirection, Nibble[bColor].Direction], bColor*15,
                        NibbleNecksRight[bDirection, Nibble[bColor].Direction], (bColor+1)*15);
  HeadImageRect := Rect(90 + 15*bDirection, bColor*15, 105 + 15*bDirection, (bColor+1)*15);

  { A prekreslime hlavu a krk. Predloha je ulozena v HiddenImage, neni
    v resources, protoze tam by mela nizsi pocet barev (resp. jinou paletu)
    a obraz by byl nekvalitni.}
  DestRect := Rect(sNewHeadX*15, sNewHeadY*15, (sNewHeadX+1)*15, (sNewHeadY+1)*15);
  GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, HeadImageRect);

  DestRect := Rect(sOldHeadX*15, sOldHeadY*15, (sOldHeadX+1)*15, (sOldHeadY+1)*15);
  GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, NeckImageRect);

  { A nakonec ulozime novy smer pohybu }
  Nibble[bColor].Direction := bDirection;

end;

{ Tato procedura prepise stav score }
procedure TGameForm.RefreshScore;
begin
  if fRefreshScore then
  begin
    BlueScoreLabel.Caption := IntToStr(Nibble[Blue].Score);
    YellowScoreLabel.Caption := IntToStr(Nibble[Yellow].Score);
    GreenScoreLabel.Caption := IntToStr(Nibble[Green].Score);
    RedScoreLabel.Caption := IntToStr(Nibble[Red].Score);
    fRefreshScore := false;
  end;
end;

{ Tato procedura vygeneruje na hraci plose novy bonus }
procedure TGameForm.GenerateNewBonus;
var
  bPosX, bPosY, bCounter1, bCounter2, bNewX, bNewY: Byte;
  fFound: Boolean;
  BonusImageRect, DestRect: TRect;
begin
  if not fGenerateNewBonus then
    Exit;

  { Volne pole jsme jeste nenasli }
  fFound := false;

  { Tremi nahodnymi pokusy zkusime najit neobsazene policko }
  for bCounter1 := 0 to 2 do
  begin
    bPosX := Random(MaxXIndex+1);
    bPosY := Random(MaxYIndex+1);
    if Desk[bPosX, bPosY] = 0 then
      begin fFound := true; Break; end;
  end;

  { Volne jsme tedy nenasli ? }
  if not fFound then
  begin
    { Nenasli, tak ho tedy najdeme sekvencne od pozice posledniho nahodneho pokusu }
    for bCounter1 := 0 to MaxXIndex do
    begin
      for bCounter2 := 0 to MaxYIndex do
      begin
        bNewX := (bPosX + bCounter1) mod (MaxXIndex+1);
        bNewY := (bPosY + bCounter2) mod (MaxYIndex+1);
        if Desk[bNewX, bNewY] = 0 then
        begin
          bPosX := bNewX;
          bPosY := bNewY;
          fFound := true;
          Break;
        end;
      end;
      if fFound then Break;
    end;
  end;

  { Volne pole uz neni = konec hry }
  if not fFound then
  begin
    fGameOver := true;
    Exit;
  end;

  { A nakonec bonus vytvorime a vykreslime }
  fGenerateNewBonus := false;
  BonusX := bPosX;
  BonusY := bPosY;
  BonusImageRect := Rect(210, 0, 225, 15);
  DestRect := Rect(bPosX*15, bPosY*15, (bPosX+1)*15, (bPosY+1)*15);
  GameImage.Canvas.CopyRect(DestRect, HiddenImage.Canvas, BonusImageRect);

end;

{ Tato procedura provede smrt hadu, tedy hlavne vymazani hada z hraci plochy }
procedure TGameForm.DoNibblesDeath;
var
  bColor, bDirection: Byte;
  wPosX, wPosY: Smallint;
  fDiedSomeone, fLiveSomeone: Boolean;
begin
  if fGameOver then
    Exit;

  { Zjistime, jestli nekdo zemrel a jestli nekdo prezil }
  fDiedSomeone := false;
  fLiveSomeone := false;
  for bColor := 0 to MaxColor do
  begin
    { Nezije jeste had? }
    if Nibble[bColor].State <> SDeath then
      begin fLiveSomeone := true; Continue; end;
    { Neni had uz odstranen ? }
    if Nibble[bColor].Length = 0 then
      Continue;
    { Had tedy zemrel nyni }
    fDiedSomeOne := true;
  end;

  { Telo prave zemreleho hada se odstrani jen kdyz jeste nektery zije }
  if fLiveSomeone and fDiedSomeone then
    for bColor := 0 to MaxColor do
    begin
      { Nezije jeste had? }
      if Nibble[bColor].State <> SDeath then
        Continue;
      { Neni had uz odstranen ? }
      if Nibble[bColor].Length = 0 then
        Continue;
      { Ted vime, ze had umrel, tak ho odstranime, zacneme od dilu pred
        ocasem, protoze ocas uz mohl byt prekreslen hlavou jineho hada. }
      wPosX := Nibble[bColor].BeforeTailX;
      wPosY := Nibble[bColor].BeforeTailY;
      while ((Desk[wPosX, wPosY] shr 30) = bColor) and (Desk[wPosX, wPosY] <> 0) do
      begin
        { Posunujeme se od ocasu k hlave a policka mazeme, na kazdem mimo
          hlavy je totiz ulozen smer k nasledujicimu. }
        GameImage.Canvas.FillRect(Rect(wPosX*15, wPosY*15, (wPosX+1)*15, (wPosY+1)*15));
        bDirection := (Desk[wPosX, wPosY] shr 28) and $03;
        Desk[wPosX, wPosY] := 0;
        wPosX := wPosX + DirectionX[bDirection];
        wPosY := wPosY + DirectionY[bDirection];
      end;

      { Nyni "vymazeme" puvodni policko ocasu z GameImage a Desk, ovsem jen pokud
        si tam jiny had uz nenamaloval hlavu }
      wPosX := Nibble[bColor].TailX;
      wPosY := Nibble[bColor].TailY;
      if (Desk[wPosX, wPosY] shr 30) = bColor then
      begin
        GameImage.Canvas.FillRect(Rect(wPosX*15, wPosY*15, (wPosX+1)*15, (wPosY+1)*15));
        Desk[wPosX, wPosY] := 0;
      end;

      { Upravime ostatni promenne }
      Nibble[bColor].Length := 0;
      Nibble[bColor].HeadX := -1;
      Nibble[bColor].HeadY := -1;
      Nibble[bColor].NewHeadX := -1;
      Nibble[bColor].NewHeadY := -1;
      Nibble[bColor].State := SDeath;
    end;

  { A pokud nekdo zemrel a hra pokracuje tak prehraj zvuk }
  if fDiedSomeone and fPlaySound and fLiveSomeone then
    sndPlaySound(DeathWave[Random(3)], SND_ASYNC);

  { Zadny had nezije = konec hry }
  if not fLiveSomeone then
    fGameOver := true;
end;

{ Tato procedura ukonci hru - vypise nejakou hlasku a predela HighScore }
procedure TGameForm.DoGameOver;
var
  bColor, bPosition, bCounter: Byte;
  Names: array[0..MaxHS] of string; { Velikost jako pocet polozek v High score }
  Scores: array[0..MaxHS] of Longint;
  NewName, OldName: string;
  NewScore, OldScore: Longint;
  fHighScoreWavePlayed: Boolean;
begin
  if not fGameOver then
    Exit;

  fHighScoreWavePlayed := false;

  { Priprava, postahujeme si dosavadni High Score }
  Timer1.Enabled := false;
  fPaused := false;
  GameOverLabel.Caption := 'GAME OVER';
  for bCounter := 0 to MaxHS do
  begin
    Names[bCounter] := MainForm.HighScoreLabels[bCounter].Name.Caption;
    Scores[bCounter] := StrToIntDef(MainForm.HighScoreLabels[bCounter].Score.Caption, 0);
  end;

  { Prepiseme High Score }
  for bColor := 0 to MaxColor do
  begin
    { Zjistujeme jmeno hrace jen pokud se umistil a pokud je to clovek }
    if (Nibble[bColor].Score <= Scores[MaxHS]) or (not Nibble[bColor].IsHuman) then
      Continue;

    { Prehrajeme zvuk pro High Score }
    if fPlaySound and (not fHighScoreWavePlayed) then
    begin
      fHighScoreWavePlayed := true;
      sndPlaySound(HighScoreWave[Random(3)], SND_ASYNC);
    end;
          
    { Ted to zjisteni jmena }
    case bColor of
      Blue:
         begin
           NameForm.NameLabel.Caption := 'Enter name of blue player:';
           NameForm.NameLabel.Font.Color := clBlue;
           NameForm.NameEdit.Font.Color := clBlue;
         end;
      Yellow:
         begin
           NameForm.NameLabel.Caption := 'Enter name of yellow player:';
           NameForm.NameLabel.Font.Color := clYellow;
           NameForm.NameEdit.Font.Color := clYellow;
         end;
      Green:
         begin
           NameForm.NameLabel.Caption := 'Enter name of green player:';
           NameForm.NameLabel.Font.Color := clLime;
           NameForm.NameEdit.Font.Color := clLime;
         end;
      Red:
         begin
           NameForm.NameLabel.Caption := 'Enter name of red player:';
           NameForm.NameLabel.Font.Color := clRed;
           NameForm.NameEdit.Font.Color := clRed;
         end;
      end; { case }
      NameForm.ShowModal;
      NewName := NameForm.NameEdit.Text;
      NewScore := Nibble[bColor].Score;

      { Zjistime, na kterou pozici umistit toto nove skore }
      bPosition := 4;
      while ((bPosition - 1) >= 0) and (Scores[bPosition-1] < NewScore) do
        Dec(bPosition);

      { A na tuto pozici skore se jmenem zapiseme, ostatni odsunem }
      for bCounter := bPosition to MaxHS do
      begin
        OldName := Names[bCounter];
        OldScore := Scores[bCounter];
        Names[bCounter] := NewName;
        Scores[bCounter] := NewScore;
        NewName := OldName;
        NewScore := OldScore;
      end;
  end; { for }

  { Prehrajeme zvuk pro Game Over }
  if fPlaySound and (not fHighScoreWavePlayed) then
    sndPlaySound(GameOverWave[Random(2)], SND_ASYNC);

  { A ulozeni }
  for bCounter := 0 to MaxHS do
  begin
    MainForm.HighScoreLabels[bCounter].Name.Caption := Names[bCounter];
    MainForm.HighScoreLabels[bCounter].Score.Caption := IntToStr(Scores[bCounter]);
  end;

end;

{ Tato funkce vygeneruje tah hada - pocitacova inteligence. Parametr bColor
  urcuje barvu hada a nabyva nasledujicich hodnot: 0=Blue, 1=Yellow, 2=Green, 3=Red. }
function TGameForm.GenerateNibbleTurn(bColor: Byte): Byte;
var
  bReturnValue: Byte;
begin
  bReturnValue := FindDirectionToSurvive(0,
                                         bColor,
                                         Nibble[bColor].Direction,
                                         Nibble[bColor].HeadX,
                                         Nibble[bColor].HeadY);
  if bReturnValue < 4 then
    GenerateNibbleTurn := bReturnValue
  else
    GenerateNibbleTurn := bReturnValue and $03;
end;

{ Tato funkce zkusi najit smer, kterym se priblizi had k bonusu a prezije
  alespon do Nibble[bColor].MaxDepth tahu (kazdy had ma svou maximalni hloubku).
  bDepth je hloubka aktualne promysleneho tahu. Zacina se na pozici [sX,sY].
  Pokud se povede prezit, funkce vraci smer kudy had muze jet, jinak nejvyznamnejsi
  bit navratove hodnoty nastavi na 1 a do 2 nejmene vyznamnych bitu ulozi smer,
  kudy mozna had prezije alespon jeste jeden tah - to proto, aby hadi "nespachali
  sebevrazdu". Parametr bDirection urcuje smer pohybu, kterym had na pole
  [sX,sY] prijel. Funkce pouziva metodu prohledavani do hloubky s castecnou heuristikou. }
function TGameForm.FindDirectionToSurvive(bDepth, bColor, bDirection: Byte; sX, sY: Shortint): Byte;
var
  { 3 nove mozne pozice, velikost poli je ale o 1 vetsi pro urychleni }
  NewPosition: array[0..3] of record
                                Direction: Byte;    { Smer kterym sem had prijel }
                                X, Y: Shortint;     { Adresa pozice }
                                Value: Shortint;    { Statisticka hodnota pozice }
                                OldDesk: Longword;  { Obsah puvodniho hraciho policka }
                                DeskPtr: ^Longword; { Ukazatel na tuto pozici
                                                      v GameDesk.Desk - pro urychleni  }
                              end;
  bIndex: Byte;
  bPossibleDirection: Byte;
  fOK: Boolean;
begin

  { Rekurze dorazila az na konec a had zije - uspech }
  if bDepth = Nibble[bColor].MaxDepth then
  begin
    { Smer na konci rekurze neni smerodatny, jen aby nebyl $FF }
    FindDirectionToSurvive := 0;
    Exit;
  end;

  { Zjistime 3 nove mozne smery pohybu, pro kazdou barvu hada ale
    trochu jinak, at jsou hadi trochu individuality. Je zajimave, jaky vliv na
    pohyb hadu maji pouze ruzne indexy pri teto inicializaci }
  if bDepth = 0 then
    bIndex := bColor mod 3
  else
    bIndex := Random(3);
  NewPosition[bIndex].Direction := bDirection;
  if bDirection <= 1 then
    begin NewPosition[(bIndex+1) mod 3].Direction := 2; NewPosition[(bIndex+2) mod 3].Direction := 3; end
  else
    begin NewPosition[(bIndex+1) mod 3].Direction := 0; NewPosition[(bIndex+2) mod 3].Direction := 1; end;

  { Nyni si pripravime promenne a kazdou novou pozici staticky ohodnotime, cim vetsi
    bude jeji hodnota, tim drive se rekurzivne vyhodnoti, upravou konstant lze
    zmenit zpusob rozhodovani hada. }
  for bIndex := 0 to 2 do
  begin
    { Ze smeru odvodime souradnice polohy }
    NewPosition[bIndex].X := sX + DirectionX[ NewPosition[bIndex].Direction ];
    NewPosition[bIndex].Y := sY + DirectionY[ NewPosition[bIndex].Direction ];

    { Ulozime ukazatel na toto policko v GameDesk.Desk - zde je dulezite, aby
      matice Desk byla o jednicku na kazde strane vetsi. }
    NewPosition[bIndex].DeskPtr := @Desk[ NewPosition[bIndex].X, NewPosition[bIndex].Y ];

    { Ulozime policko puvodni hraci plochy }
    NewPosition[bIndex].OldDesk := NewPosition[bIndex].DeskPtr^;

    { Toto se provadi jen pri zakladnim rozhodovani }
    if bDepth = 0 then
    begin
      { Inicializujeme statickou hodnotu - podle inteligence hada uplatnime nahodu, zde
        je dulezite, aby MaxDepth[bColor] bylo mensi nez 34. }
      NewPosition[bIndex].Value := 32 + Random(25 - (Nibble[bColor].MaxDepth div 2));

      { Ohodnoceni podle priblizeni se k bonusu }
      if ( Abs(NewPosition[bIndex].X-BonusX) < Abs(sX-BonusX) )
         or ( Abs(NewPosition[bIndex].Y-BonusY) < Abs(sY-BonusY) ) then
        Inc(NewPosition[bIndex].Value, 14);

      { Je toto policko ocasem nektereho hada ? - Pokud had sezere bonus, tento
        by zemrel. }
      if ((Nibble[NewPosition[bIndex].OldDesk shr 30].Length
           + (NewPosition[bIndex].OldDesk and $00FFFFFF)) = (dwGameTurn + bDepth))
         and (NewPosition[bIndex].OldDesk shr 30 <> bColor) then
        Dec(NewPosition[bIndex].Value, 16);

      { Je toto policko vedle hlavy nejakeho zijiciho hada, ze by se mohli srazit ?
        Zde je dulezite, aby souradnice hlavy nehrajiciho hada byla [-1,-1]. }
      if (Abs(Nibble[(bColor+1) mod TotColors].HeadX - NewPosition[bIndex].X)
          + Abs(Nibble[(bColor+1) mod TotColors].HeadY - NewPosition[bIndex].Y) <= 1)
         or (Abs(Nibble[(bColor+2) mod TotColors].HeadX - NewPosition[bIndex].X)
             + Abs(Nibble[(bColor+2) mod TotColors].HeadY - NewPosition[bIndex].Y) <= 1)
         or (Abs(Nibble[(bColor+3) mod TotColors].HeadX - NewPosition[bIndex].X)
             + Abs(Nibble[(bColor+3) mod TotColors].HeadY - NewPosition[bIndex].Y) <= 1) then
         Dec(NewPosition[bIndex].Value, 16);
    end
    else { bDepth je vetsi nez 0, tedy staticke ohodnoceni uz nepouzivame, nedavalo
           by totiz dobre vysledky. }
      NewPosition[bIndex].Value := 0;

    { Neni tato pozice oznacena jako jiz prohledana (tedy ma 5. nejvyznamnejsi bit = 1) ? -
      Je-li, nastavime Value tak, aby se vubec nezkousela. }
    if (NewPosition[bIndex].OldDesk and $08000000) > 0 then
      Dec(NewPosition[bIndex].Value, 80);

  end; { for }

  { Toto take provedeme pouze pri zakladnim rozhodovani }
  if bDepth = 0 then
  begin
    { Setridime pozice podle jejich vyhodnosti, tj. podle vyssi Value = heuristika.
      Misto nejakych for cyklu pouzivame principu transpozicni tridici site
      (kombinacniho obvodu). Vyuzivame zde tez toho, ze pole s pozicemi ma
      o 1 vetsi velikost = pomocna promenna }
    { Je-li 1. pozice vyhodnejsi nez 0., prohodime je }
    if NewPosition[0].Value < NewPosition[1].Value then
    begin
      NewPosition[3] := NewPosition[0];
      NewPosition[0] := NewPosition[1];
      NewPosition[1] := NewPosition[3];
    end;
    { Je-li 2. pozice vyhodnejsi nez 0., prohodime je }
    if NewPosition[0].Value < NewPosition[2].Value then
    begin
      NewPosition[3] := NewPosition[0];
      NewPosition[0] := NewPosition[2];
      NewPosition[2] := NewPosition[3];
    end;
    { Je-li 2. pozice vyhodnejsi nez 1., prohodime je }
    if NewPosition[1].Value < NewPosition[2].Value then
    begin
      NewPosition[3] := NewPosition[2];
      NewPosition[2] := NewPosition[1];
      NewPosition[1] := NewPosition[3];
    end;

  end;

  { Oznacime vsechny tri pozice jako jiz prohledane - 3x stejny kod }
  NewPosition[0].DeskPtr^ := NewPosition[0].DeskPtr^ or $08000000;
  NewPosition[1].DeskPtr^ := NewPosition[1].DeskPtr^ or $08000000;
  NewPosition[2].DeskPtr^ := NewPosition[2].DeskPtr^ or $08000000;

  { Nyni provedeme dalsi rekurzivni vnoreni }
  bIndex := 0;
  fOK := false;
  bPossibleDirection := 0;
  while (not fOK) and (bIndex <= 2) do
  begin
    { Promenne sX a sY uz nepotrebujeme, tak je vyuzijeme }
    sX := NewPosition[bIndex].X;
    sY := NewPosition[bIndex].Y;

    { Prezije had na teto pozici a vyplati se ji zkusit ? }
    if (Nibble[NewPosition[bIndex].OldDesk shr 30].Length
         + (NewPosition[bIndex].OldDesk and $00FFFFFF) <= dwGameTurn + bDepth)
       and (sY >= 0) and (sY <= MaxYIndex) and (sX >= 0) and (sX <= MaxXIndex)
       and (NewPosition[bIndex].Value >= 0) then
    begin
      { Do hraci desky ulozime tento pohyb }
      NewPosition[bIndex].DeskPtr^ := (Longword(bColor) shl 30) or $08000000 or (dwGameTurn + bDepth);

      { Ulozime si i smer prijezdu na toto policko }
      bPossibleDirection := NewPosition[bIndex].Direction;

      if FindDirectionToSurvive(bDepth+1, bColor, NewPosition[bIndex].Direction, sX, sY) <= MaxDirection then
        fOK := true;
    end;

    Inc(bIndex);
  end; { while }

  { Obnovime hraci desku do puvodniho stavu - 3x stejny kod }
  NewPosition[0].DeskPtr^ := NewPosition[0].OldDesk;
  NewPosition[1].DeskPtr^ := NewPosition[1].OldDesk;
  NewPosition[2].DeskPtr^ := NewPosition[2].OldDesk;

  { A nyni jiz zname vysledky, tak se podle nich zaridime }
  if fOk then
    FindDirectionToSurvive := bPossibleDirection
  else
    FindDirectionToSurvive := $80 or bPossibleDirection;

end;


{ Procedura nastane pri stisku ESC, pausuje hru }
procedure TGameForm.PauseButtonClick(Sender: TObject);
begin
  if fGameOver then
    Exit;
  if fPaused then
  begin
    Timer1.Enabled := true;
    fPaused := false;
    GameOverLabel.Caption := '';
    TitleLabel.Caption := '';
  end
  else
  begin
    Timer1.Enabled := false;
    fPaused := true;
    GameOverLabel.Caption := 'PAUSED';
    TitleLabel.Caption := '[ Press ESC to continue ]';
  end;

end;

{ Procedura nastavajici pri uzavreni formy }
procedure TGameForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  { Uklidime po sobe }
  Timer1.Enabled := false;
  if not fGameOver then
  begin
    fGameOver := true;
    DoGameOver;
  end;
end;

end.
